<?php

namespace WooCommerce\FedEx;

include_once WC_SHIPPING_FEDEX_API_DIR . '/abstract/class-abstract-fedex-api.php';

class FedEx_Legacy_API extends Abstract_FedEx_API {

	/**
	 * Endpoint for the FedEx Rating API.
	 *
	 * @var string
	 */
	protected $endpoint = '';

	/**
	 * Web Services Key.
	 *
	 * @var string
	 */
	private $api_key;

	/**
	 * Web Services Secret.
	 *
	 * @var string
	 */
	private $api_pass;

	/**
	 * Web Services Secret.
	 *
	 * @var bool
	 */

	/**
	 * FedEx Account Number
	 *
	 * @var string
	 */
	private $account_number;

	/**
	 * FedEx Account Meter
	 *
	 * @var string
	 */
	private $meter_number;

	private $api_mode;
	/**
	 * Is Shipper Address Residential.
	 *
	 * @var bool
	 */

	/**
	  * Request type
	  *
	  * @var string
	  */
	private $request_type;
	/**
	 * Is Recipient Address Residential.
	 *
	 * @var bool
	 */
	private $residential;

	static $rate_service_version = 31;
	static $address_validation_service_version = 4;

	/**
	 * @param \WC_Shipping_FedEx $fedex_shipping_method The FedEx shipping method object.
	 *
	 */
	public function __construct( $fedex_shipping_method ) {
		$this->shipping_method = $fedex_shipping_method;
		$this->api_mode        = $fedex_shipping_method->is_production()  ? 'production' : 'test';
		$this->api_key         = $fedex_shipping_method->get_api_key();
		$this->api_pass        = $fedex_shipping_method->get_api_pass();
		$this->account_number  = $fedex_shipping_method->get_account_number();
		$this->meter_number    = $fedex_shipping_method->get_meter_number();
		$this->request_type    = $fedex_shipping_method->get_request_type();
		$this->residential     = $fedex_shipping_method->is_residential();
		$this->logger          = wc_get_logger();
	}

	/**
	 * get_fedex_api_request function
	 *
	 * @param string $request_type
	 *
	 * @return array
	 */
	public function get_fedex_api_request( $request_type ) {
		$request['WebAuthenticationDetail'] = array(
			'UserCredential' => array(
				'Key'      => $this->api_key,
				'Password' => $this->api_pass,
			),
		);

		$request['ClientDetail']            = array(
			'AccountNumber' => $this->account_number,
			'MeterNumber'   => $this->meter_number,
		);

		$request['TransactionDetail']       = array(
			'CustomerTransactionId' => ' *** WooCommerce Rate Request ***',
		);

		$request['Version']                 = array(
			'ServiceId'    => 'crs',
			'Major'        => self::$rate_service_version,
			'Intermediate' => '0',
			'Minor'        => '0',
		);

		// Include PREFERRED Rate Type.
		// The preferred currency is not returned if the requested currency is already present within the rate response.
		$request['RequestedShipment']['RateRequestTypes'] = 'LIST' === $this->request_type ? array( 'LIST', 'PREFERRED' ) : 'PREFERRED';

		$request['RequestedShipment']['DropoffType'] = 'REGULAR_PICKUP';

		return $request;
	}

	/**
	 * Get result.
	 *
	 * @param array  $request
	 * @param string $request_type
	 * @return array
	 */
	public function get_result( $request, $request_type ) {

		$rate_soap_file_location = plugin_dir_path( dirname( __FILE__ ) ) . 'legacy/' . $this->api_mode . '/RateService_v' . self::$rate_service_version . '.wsdl';

		try {

			$client = new \SoapClient( $rate_soap_file_location, array( 'trace' => 1 ) );
			$result = $client->getRates( $request );
		} catch ( \Exception $e ) {

			$stream_context_args = array( 'ssl' => array( 'verify_peer' => false, 'verify_peer_name' => false ) );
			$soap_args           = array(
				'trace'          => 1,
				'stream_context' => stream_context_create( $stream_context_args ),
			);

			$client = new \SoapClient( $rate_soap_file_location, $soap_args );
			$result = $client->getRates( $request );
		}

		return $result;
	}

	/**
	 * Check if Destination Address is Residential.
	 *
	 * @param array $request
	 * @return bool
	 */
	public function residential_address_validation( $request ) {
		$request['Version'] = array(
			'ServiceId'    => 'aval',
			'Major'        => self::$address_validation_service_version,
			'Intermediate' => '0',
			'Minor'        => '0',
		);

		$request['WebAuthenticationDetail'] = array(
			'UserCredential' => array(
				'Key'      => $this->api_key,
				'Password' => $this->api_pass,
			)
		);

		$request['ClientDetail'] = array(
			'AccountNumber' => $this->account_number,
			'MeterNumber'   => $this->meter_number
		);

		$request['TransactionDetail']       = array( 'CustomerTransactionId' => ' *** Address Validation Request v' . self::$address_validation_service_version . ' from WooCommerce ***' );

		// Check if address is residential or commerical
		try {

			$client = new \SoapClient( plugin_dir_path( dirname( __FILE__ ) ) . 'legacy/' . $this->api_mode . '/AddressValidationService_v' . self::$address_validation_service_version . '.wsdl', array( 'trace' => 1 ) );

			$response = $client->addressValidation( $request );

			if ( $response->HighestSeverity !== 'SUCCESS' ) {
				return $this->residential;
			}

			if ( is_array( $response->AddressResults ) ) {
				$addressResult = $response->AddressResults[0];
			} else {
				$addressResult = $response->AddressResults;
			}

			return ( isset( $addressResult->Classification ) && 'RESIDENTIAL' === $addressResult->Classification ) ? true : $this->residential;

		} catch ( \Exception $e ) {
		}

 		return $this->residential;
	}

	public function process_result( $result = '' ) {
		if ( ! $result || empty ( $result->RateReplyDetails ) ) {
			return;
		}

		$rate_reply_details = $result->RateReplyDetails;
		$store_currency     = get_woocommerce_currency();

		// Workaround for when an object is returned instead of array
		if ( is_object( $rate_reply_details ) && isset( $rate_reply_details->ServiceType ) ) {
			$rate_reply_details = array( $rate_reply_details );
		}

		if ( ! is_array( $rate_reply_details ) ) {
			return;
		}

		foreach ( $rate_reply_details as $quote ) {

			if ( is_array( $quote->RatedShipmentDetails ) ) {

				$details = false;
				foreach ( $quote->RatedShipmentDetails as $i => $d ) {
					if (
						// Attempt to return Rate Type set in Shipping Zone settings, ACCOUNT or LIST.
						strstr( $d->ShipmentRateDetail->RateType, 'PAYOR_' . $this->shipping_method->get_request_type() ) &&
						$d->ShipmentRateDetail->TotalNetCharge->Currency === $store_currency
					) {
						$details = $quote->RatedShipmentDetails[ $i ];
						break;
					}
				}

				// If first attempt don't return Shipment Details due to currency mismatch, look for PREFERRED rate.
				if ( ! $details ) {
					foreach ( $quote->RatedShipmentDetails as $i => $d ) {
						if (
							strstr( $d->ShipmentRateDetail->RateType, 'PREFERRED_' . $this->shipping_method->get_request_type() ) &&
							$d->ShipmentRateDetail->TotalNetCharge->Currency === $store_currency
						) {
							$details = $quote->RatedShipmentDetails[ $i ];
							break;
						}
					}
				}
			} else {
				$details = $quote->RatedShipmentDetails;
			}

			if ( empty( $details ) ) {
				continue;
			}

			$currency  = isset( $details->ShipmentRateDetail->TotalNetCharge->Currency ) ? $details->ShipmentRateDetail->TotalNetCharge->Currency : '';
			$rate_name = strval( $this->shipping_method->get_service( $quote->ServiceType ) );

			if ( apply_filters( 'woocommerce_shipping_fedex_check_store_currency', true, $currency, $details, $this->shipping_method ) && ( $store_currency !== $currency ) ) {
				/* translators: 1: FedEx service name 2: currency for the rate 3: store's currency */
				$this->shipping_method->debug( sprintf( __( '[FedEx] Rate for %1$s is in %2$s but store currency is %3$s.', 'woocommerce-shipping-fedex' ), $rate_name, $currency, $store_currency ) );
				continue;
			}

			$rate_code = strval( $quote->ServiceType );
			$rate_id   = $this->shipping_method->get_rate_id( $rate_code );
			$rate_cost = floatval( $details->ShipmentRateDetail->TotalNetCharge->Amount );

			$this->shipping_method->prepare_rate( $rate_code, $rate_id, $rate_name, $rate_cost, $currency, $details );
		}
	}

}
