<?php

namespace WooCommerce\FedEx;

/**
 * Class FedEx_OAuth
 *
 * Handles the OAuth authentication for the FedEx REST API.
 *
 * @package WooCommerce\FedEx
 */
class FedEx_OAuth {

	/**
	 * The client ID used to authenticate with the FedEx OAuth API.
	 *
	 * @var string
	 */
	private $client_id;
	/**
	 * The client secret used to authenticate with the FedEx OAuth API.
	 *
	 * @var string
	 */
	private $client_secret;
	/**
	 * The URL for the FedEx OAuth API endpoint.
	 *
	 * @var string
	 */
	private $endpoint;
	/**
	 * The name of the transient used to store the access token.
	 *
	 * @var string
	 */
	private $access_token_transient_name = 'woocommerce_fedex_oauth_access_token';

	/**
	 * Constructor.
	 *
	 * @param object $fedex_shipping_method The FedEx shipping method object.
	 */
	public function __construct( $fedex_shipping_method ) {
		$this->client_id     = $fedex_shipping_method->get_client_id();
		$this->client_secret = $fedex_shipping_method->get_client_secret();
		$this->endpoint      = $fedex_shipping_method->is_production() ? 'https://apis.fedex.com/oauth/token' : 'https://apis-sandbox.fedex.com/oauth/token';
	}

	/**
	 * Check if we've successfully authenticated.
	 *
	 * @return bool
	 */
	public function is_authenticated() {
		return (bool) $this->get_access_token();
	}

	/**
	 * Get an access token.
	 *
	 * @return string|bool
	 */
	public function get_access_token() {
		$access_token = get_transient( $this->access_token_transient_name );

		if ( false === $access_token ) {

			$response = $this->get_access_token_from_fedex();
			if ( $response && ! empty( $response->access_token ) && ! empty( $response->expires_in ) ) {
				set_transient( $this->access_token_transient_name, $response->access_token, $response->expires_in );

				$access_token = $response->access_token;
			}
		}

		return $access_token;
	}

	/**
	 * Get an access token from the FedEx OAuth API.
	 *
	 * @return object|bool The response from the FedEx OAuth API, or false if the request fails.
	 */
	private function get_access_token_from_fedex() {
		$headers = array(
			'Content-Type' => 'application/x-www-form-urlencoded'
		);

		$body = array(
			'grant_type'    => 'client_credentials',
			'client_id'     => $this->client_id,
			'client_secret' => $this->client_secret
		);

		$response = wp_remote_post( $this->endpoint, array(
			'headers' => $headers,
			'body'    => $body
		) );

		if ( is_wp_error( $response ) ) {
			$error_message = $response->get_error_message();

			$logger = wc_get_logger();
			$logger->error( "FedEx_OAuth::request_access_token: The FedEx OAuth endpoint returned the following error: $error_message" );

			return false;
		}

		return json_decode( wp_remote_retrieve_body( $response ) );
	}

}
