<?php

namespace WooCommerce\FedEx;

include_once WC_SHIPPING_FEDEX_API_DIR . '/abstract/class-abstract-fedex-api.php';

use WC_Shipping_FedEx;

class FedEx_REST_API extends Abstract_FedEx_API {

	/**
	 * Endpoint for the FedEx Rating API.
	 *
	 * @var string
	 */
	protected $api_url = '';
	/**
	 * @var \WooCommerce\FedEx\FedEx_OAuth
	 */
	private $oauth;
	/**
	 * FedEx Account Number
	 *
	 * @var string
	 */
	private $account_number;
	/**
	 * Request type
	 *
	 * @var string
	 */
	private $request_type;
	/**
	 * Is Recipient Address Residential.
	 *
	 * @var bool
	 */
	private $residential;
	/**
	 * Request headers.
	 *
	 * @var array
	 */
	private $headers;

	/**
	 * FedEx_REST_API constructor.
	 *
	 * @param WC_Shipping_FedEx $fedex_shipping_method The FedEx shipping method object.
	 */
	public function __construct( $fedex_shipping_method ) {
		$this->shipping_method = $fedex_shipping_method;
		$this->account_number  = $fedex_shipping_method->get_account_number();
		$this->request_type    = $fedex_shipping_method->get_request_type();
		$this->oauth           = $fedex_shipping_method->get_oauth();
		$this->residential     = $fedex_shipping_method->is_residential();
		$this->api_url         = $fedex_shipping_method->is_production() ? 'https://apis.fedex.com' : 'https://apis-sandbox.fedex.com';
		$this->headers         = array(
			'Authorization' => 'Bearer ' . $this->oauth->get_access_token(),
			'X-locale'      => 'en_US',
			'Content-Type'  => 'application/json',
		);
	}

	/**
	 * Get the common API request parameters.
	 *
	 * @param string $request_type
	 *
	 * @return array
	 */
	public function get_fedex_api_request( $request_type ) {
		$request['AccountNumber'] = $this->account_number;

		$request['rateRequestControlParameters'] = array(
			'returnTransitTimes'          => false,
			'servicesNeededOnRateFailure' => true,
			'rateSortOrder'               => 'SERVICENAMETRADITIONAL',
		);

		// Include PREFERRED Rate Type.
		// The preferred currency is not returned if the requested currency is already present within the rate response.
		$request['RequestedShipment']['RateRequestType'] = array( $this->request_type, 'PREFERRED' );

		if ( 'freight' !== $request_type ) {
			$request['RequestedShipment']['PickupType'] = 'DROPOFF_AT_FEDEX_LOCATION';
			$request['carrierCodes']                    = array( 'FDXE', 'FDXG', 'FXSP' );
		}

		if ( 'default' === $request_type ) {
			$request['carrierCodes'] = array( 'FDXE', 'FDXG', 'FXSP' );
		}

		if ( 'ground' === $request_type ) {
			$request['carrierCodes'] = array( 'FDXG' );
		}

		if ( 'smartpost' === $request_type ) {
			$request['carrierCodes'] = array( 'FXSP' );
		}

		return $request;
	}

	/**
	 * Get result.
	 *
	 * @param array  $request
	 * @param string $request_type
	 *
	 * @return array
	 */
	public function get_result( $request, $request_type ) {

		$args = array(
			'headers' => $this->headers,
			'body'    => json_encode( $request ),
		);

		$endpoint = 'freight' === $request_type ? '/rate/v1/freight/rates/quotes' : '/rate/v1/rates/quotes';

		$response      = wp_remote_post( $this->api_url . $endpoint, $args );
		$response_body = wp_remote_retrieve_body( $response );

		return json_decode( $response_body );
	}

	/**
	 * Check if Destination Address is Residential.
	 *
	 * @param array $request
	 *
	 * @return bool
	 */
	public function residential_address_validation( $request ) {
		$request_body = $this->prepare_request( $request );

		$args = array(
			'headers' => $this->headers,
			'body'    => json_encode( $request_body ),
		);

		$response = wp_remote_post( $this->api_url . '/address/v1/addresses/resolve', $args );

		$response_body = wp_remote_retrieve_body( $response );
		$result        = json_decode( $response_body );

		return ( isset( $result->output->resolvedAddresses[0]->classification ) && 'RESIDENTIAL' === $result->output->resolvedAddresses[0]->classification ) ? true : $this->residential;
	}

	/**
	 * Prepare request elements.
	 *
	 * Depending on the request type, some elements need to be renamed or removed.
	 * This function will loop through the request array, and modify the keys and values as needed.
	 *
	 * @param array  $request      The request array.
	 * @param string $request_type The request type.
	 * @param string $parent_key   The parent key.
	 *
	 * @return array
	 */
	public function prepare_request( $request, $request_type = '', $parent_key = '' ) {

		if ( ! is_array( $request ) ) {
			return array();
		}

		$updated_request = array();
		foreach ( $request as $key => $value ) {
			$original_key = $key;

			if ( 'AccountNumber' === $key ) {
				$value = array( 'value' => $value );
			}

			if ( 'PackageCount' === $key ) {
				$key = 'totalPackageCount';
			}

			if ( 'freight' === $request_type && 'PackagingType' === $key ) {
				continue;
			}

			if ( 'freight' === $request_type && 'RequestedShipment' === $key ) {
				$key = 'freightRequestedShipment';
			}

			if ( 'freight' === $request_type && 'PhysicalPackaging' === $key ) {
				$key = 'subPackagingType';
			}

			if ( 'freight' === $request_type && 'FedExFreightAccountNumber' === $key ) {
				$key   = 'accountNumber';
				$value = array( 'value' => $value );
			}

			if ( 'freight' === $request_type && 'RequestedPackageLineItems' === $key ) {
				$value = array( $value );
			}

			if ( 'freight' === $request_type && 'LineItems' === $key ) {
				$key   = 'lineItem';
				$value = array( $value );
			}

			if ( 'freight' === $request_type && 'Packaging' === $key ) {
				$key = 'subPackagingType';
			}

			if ( 'freight' === $request_type && 'LineItems' === $parent_key && 'Id' === $key ) {
				$value = (string) $value;
			}

			if ( 'freight' === $request_type && 'AssociatedFreightLineItems' === $key ) {
				$value = array( $value );
			}

			if ( 'SmartPostDetail' === $key ) {
				$key = 'smartPostInfoDetail';
			}

			if ( 'ShipTimestamp' === $key ) {
				$value = date( 'Y-m-d', strtotime( $value ) );
				$key   = 'shipDateStamp';
			}

			if ( ( 'freight' !== $request_type && 'ShippingChargesPayment' === $key ) || 'SequenceNumber' === $key || 'GroupNumber' === $key || 'DropoffType' === $key || '' === $value ) {
				continue;
			}

			$updated_request[ lcfirst( $key ) ] = is_array( $value ) ? $this->prepare_request( $value, $request_type, $original_key ) : $value;
		}

		return $updated_request;
	}

	/**
	 * Process result.
	 *
	 * @param object $result
	 *
	 * @return void
	 */
	public function process_result( $result ) {
		if ( ! isset( $result->output ) || ! isset( $result->output->rateReplyDetails ) ) {
			return;
		}

		$store_currency = get_woocommerce_currency();

		foreach ( $result->output->rateReplyDetails as $quote ) {

			// Prevent double rates we should already have rate for "FedEx SmartPost®" both rates have same service key "SMART_POST"
			if ( isset( $quote->serviceName ) && 'FedEx SmartPost Bound Print Matter®' === $quote->serviceName ) {
				continue;
			}

			if ( is_array( $quote->ratedShipmentDetails ) ) {
				$details = false;
				foreach ( $quote->ratedShipmentDetails as $i => $d ) {
					// Attempt to return Rate Type set in Shipping Zone settings, ACCOUNT or LIST.
					if (
						strstr( $d->rateType, $this->shipping_method->get_request_type() ) &&
						$d->currency === $store_currency
					) {
						$details = $quote->ratedShipmentDetails[ $i ];
						break;
					}
				}

				// If first attempt don't return Shipment Details due to currency mismatch, look for PREFERRED rate.
				if ( ! $details ) {
					foreach ( $quote->ratedShipmentDetails as $i => $d ) {
						if (
							strstr( $d->rateType, 'PREFERRED_CURRENCY' ) &&
							$d->currency === $store_currency
						) {
							$details = $quote->ratedShipmentDetails[ $i ];
							break;
						}
					}
				}
			} else {
				$details = $quote->ratedShipmentDetails;
			}

			if ( empty( $details ) ) {
				continue;
			}

			$currency  = isset( $details->currency ) ? $details->currency : '';
			$rate_name = strval( $this->shipping_method->get_service( $quote->serviceType ) );

			if ( apply_filters( 'woocommerce_shipping_fedex_check_store_currency', true, $currency, $details, $this->shipping_method ) && ( $store_currency !== $currency ) ) {
				/* translators: 1: FedEx service name 2: currency for the rate 3: store's currency */
				$this->shipping_method->debug( sprintf( __( '[FedEx] Rate for %1$s is in %2$s but store currency is %3$s.', 'woocommerce-shipping-fedex' ), $rate_name, $currency, $store_currency ) );
				continue;
			}

			$rate_code = strval( $quote->serviceType );
			$rate_id   = $this->shipping_method->get_rate_id( $rate_code );
			$rate_cost = floatval( $details->totalNetCharge );

			$this->shipping_method->prepare_rate( $rate_code, $rate_id, $rate_name, $rate_cost, $currency, $details );
		}
	}

}
